#include "IZ_app.h"

IZ_ProcedureResult IZ_AppWSClientInitialize(IZ_App* app, IZ_WSClientInitializeParams params) {
	if (IZ_WSClientInitialize(&app->client, params)) {
		printf("error\n");
		return -1;
	}
	return 0;
}

IZ_ProcedureResult IZ_AppConnect(void* app_raw) {
	IZ_App* app = app_raw;
	IZ_WSClientInitializeParams params = {
		.address = "127.0.0.1",
		.path = "/",
		.port = 42069,
		.userdata = app,
	};

	if (IZ_AppWSClientInitialize(app, params)) {
		return -1;
	}

	i32 result = 0;
	while (true) {
		if (IZ_WSClientHandle(&app->client)) {
			result = -1;
			break;
		}

		if (app->client.ws.interrupted) {
			break;
		}
	}

	IZ_WSClientTeardown(&app->client);
	return result;
}

void IZ_AppEstablishConnection(IZ_App* app) {
	if (app->client.ws.context) {
		return;
	}

	app->client_thread = SDL_CreateThread(IZ_AppConnect, "networking", app);
	SDL_DetachThread(app->client_thread);
}

void IZ_AppCloseConnection(IZ_App* app) {
	if (!app->client.ws.context) {
		return;
	}
	app->client.ws.interrupted = true;
	IZ_WSClientCancelService(&app->client);
}

IZ_ProcedureResult IZ_AppInitialize(IZ_App* app, u8 argc, const char* argv[]) {
	memset(app, 0, sizeof(IZ_App));
	u32 flags = (
		SDL_INIT_VIDEO
		| SDL_INIT_GAMECONTROLLER
		| SDL_INIT_EVENTS
	);

	if (SDL_Init(flags) < 0) {
		SDL_LogError(SDL_LOG_CATEGORY_ERROR, "SDL could not initialize! SDL_Error: %s\n", SDL_GetError());
		return IZ_APP_RUN_SDL_INIT_ERROR;
	}

	const char* cmdline_buffer;
	char config_path[128];
	// TODO abstract command line args parsing
	if ((cmdline_buffer = lws_cmdline_option(argc, argv, "-c"))) {
		memcpy_s(config_path, 128, cmdline_buffer, 128);
	} else {
		IZ_ConfigGetDefaultPath(config_path, 128);
	}

	if (IZ_VideoInitialize(&app->video_state, config_path, argc, argv)) {
		return IZ_APP_RUN_VIDEO_INIT_ERROR;
	}

	if (IZ_InputInitialize(&app->input_state, config_path, argc, argv)) {
		return IZ_APP_RUN_INPUT_INIT_ERROR;
	}

	IZ_PoolInitialize(&app->pool, POOL_MAX_SIZE);

  // TODO put into its timer module
	app->ticks = 0;
	return IZ_APP_RUN_RESULT_OK;
}

void IZ_AppTeardown(IZ_App* app) {
	IZ_AppCloseConnection(app);
	IZ_PoolTeardown(&app->pool);
	IZ_InputTeardown(&app->input_state);
	IZ_VideoTeardown(&app->video_state);
	SDL_Quit();
}

IZ_ProcedureResult IZ_AppHandleSDLEvents(IZ_App* app) {
	SDL_Event e;
	while (SDL_PollEvent(&e) != 0) {
		if (e.type == SDL_QUIT) {
			return 1;
		}

		if (e.type == SDL_KEYDOWN) {
			if (e.key.keysym.sym == SDLK_PAGEUP) {
				IZ_AppEstablishConnection(app);
			} else if (e.key.keysym.sym == SDLK_PAGEDOWN) {
				IZ_AppCloseConnection(app);
			}
		}

		IZ_InputHandleSDLEvents(&app->input_state, e);
	}
	return 0;
}

void IZ_AppHandlePortMIDIEvents(IZ_App* app) {
	u8 player_index;
	i32* midi_events_count;
	u32 midi_event_index;
	for (player_index = 0; player_index < PLAYERS; player_index += 1) {
		if (!app->input_state.midi_input_state[player_index].device_info) {
			continue;
		}

		midi_events_count = &app->input_state.midi_input_state[player_index].midi_events_count;
		*midi_events_count = Pm_Read(
			app->input_state.midi_input_state[player_index].stream,
			// TODO bind buffers and streams to device instead of player input state
			app->input_state.midi_input_state[player_index].event_buffer,
			1024
		);

		if (*midi_events_count < 1) {
			continue;
		}

		for (midi_event_index = 0; midi_event_index < *midi_events_count; midi_event_index += 1) {
			IZ_InputHandlePortMIDIEvents(
				&app->input_state,
				app->input_state.midi_input_state[player_index].event_buffer[midi_event_index]
			);
		}
	}
}

void IZ_AppHandleNetworkingInboundBinaryEvents(IZ_App* app, void* binary_raw, size_t len) {
	u8* binary = binary_raw;
	size_t i;
	printf("Binary");
	for (i = 0; i < len; i += 1) {
		printf("%c%x", i == 0 ? '(' : ' ', binary[i]);
	}
	printf(")\n");
}

void IZ_AppHandleNetworkingInboundTextEvents(IZ_App* app, const char* text, size_t len) {
	printf("String(%s)\n", text);
}

IZ_ProcedureResult IZ_AppHandleInputEvents(IZ_App* app) {
	i32 sdl_events_result = IZ_AppHandleSDLEvents(app);
	if (sdl_events_result) {
		return sdl_events_result;
	}

	IZ_AppHandlePortMIDIEvents(app);
	return 0;
}

IZ_ProcedureResult IZ_AppRun(IZ_App* app, u8 argc, const char* argv[]) {
	IZ_ProcedureResult init_result = IZ_AppInitialize(app, argc, argv);
	if (init_result) {
		return init_result;
	}

	while (true) {
		app->ticks = SDL_GetTicks64();

		// TODO do audio processing
		// TODO do networking
		//if (app->client.connection) {
			// FIXME stuck in infinite loop
			//IZ_WSClientHandle(&app->client);
		//}

		if (IZ_AppHandleInputEvents(app)) {
			break;
		}

		IZ_VideoUpdate(&app->video_state, app->ticks, &app->input_state);
	}

	IZ_AppTeardown(app);
	return IZ_APP_RUN_RESULT_OK;
}

void IZ_WSClientAttemptConnect(struct lws_sorted_usec_list *sul) {
	IZ_WSClientVHostData* vhd = lws_container_of(sul, IZ_WSClientVHostData, sul);

	vhd->i.context = vhd->context;
	vhd->i.port = *vhd->port;
	vhd->i.address = vhd->address;
	vhd->i.path = vhd->path;
	vhd->i.host = vhd->i.address;
	vhd->i.origin = vhd->i.address;
	vhd->i.ssl_connection = 0;

	vhd->i.protocol = NETWORK_PROTOCOL;
	vhd->i.pwsi = &vhd->client_wsi;

	if (lws_client_connect_via_info(&vhd->i)) {
		return;
	}

	lws_sul_schedule(
		vhd->context,
		0,
		&vhd->sul,
		IZ_WSClientAttemptConnect,
		10 * LWS_US_PER_SEC
	);
}

IZ_ProcedureResult IZ_WSClientProtocolInitialize(struct lws* wsi, void* in) {
	IZ_WSClientVHostData* vhd_instance = (IZ_WSClientVHostData*) lws_protocol_vh_priv_get(
		lws_get_vhost(wsi),
		lws_get_protocol(wsi)
	);
	IZ_WSClientVHostData** vhd = &vhd_instance;
	*vhd = lws_protocol_vh_priv_zalloc(
		lws_get_vhost(wsi),
		lws_get_protocol(wsi),
		sizeof(IZ_WSClientVHostData)
	);
	(*vhd)->context = lws_get_context(wsi);
	(*vhd)->protocol = lws_get_protocol(wsi);
	(*vhd)->vhost = lws_get_vhost(wsi);
	(*vhd)->port = (u16*) lws_pvo_search(
		(const struct lws_protocol_vhost_options *)in,
		"port"
	)->value;
	(*vhd)->address = lws_pvo_search(
		(const struct lws_protocol_vhost_options *)in,
		"address"
	)->value;
	(*vhd)->path = lws_pvo_search(
		(const struct lws_protocol_vhost_options *)in,
		"path"
	)->value;
	(*vhd)->app = lws_pvo_search(
		(const struct lws_protocol_vhost_options *)in,
		"app"
	)->value;
	(*vhd)->ring = lws_ring_create(
		sizeof(IZ_WebsocketMessage),
		RING_COUNT,
		IZ_WebsocketDestroyMessage
	);
	if (!(*vhd)->ring) {
		return -1;
	}
	IZ_WSClientAttemptConnect(&(*vhd)->sul);
	return 0;
}

void IZ_WSClientProtocolTeardown(struct lws* wsi) {
	IZ_WSClientVHostData* vhd = (IZ_WSClientVHostData*) lws_protocol_vh_priv_get(
		lws_get_vhost(wsi),
		lws_get_protocol(wsi)
	);

	vhd->finished = true;
	if (vhd->ring) {
		lws_ring_destroy(vhd->ring);
	}

	lws_sul_cancel(&vhd->sul);
}

void IZ_WSClientConnectionError(struct lws* wsi, void* in) {
	IZ_WSClientVHostData* vhd = (IZ_WSClientVHostData*) lws_protocol_vh_priv_get(
		lws_get_vhost(wsi),
		lws_get_protocol(wsi)
	);

	lwsl_err("CLIENT_CONNECTION_ERROR: %s\n", in ? (char *)in : "(null)");
	vhd->client_wsi = NULL;
	lws_sul_schedule(
		vhd->context,
		0,
		&vhd->sul,
		IZ_WSClientAttemptConnect,
		LWS_US_PER_SEC
	);
}

IZ_ProcedureResult IZ_WSClientOnOpen(struct lws* wsi, IZ_WSClientSessionData* pss) {
	IZ_WSClientVHostData* vhd = (IZ_WSClientVHostData*) lws_protocol_vh_priv_get(
		lws_get_vhost(wsi),
		lws_get_protocol(wsi)
	);

	vhd->established = true;
	pss->ring = lws_ring_create(sizeof(IZ_WebsocketMessage), RING_COUNT,IZ_WebsocketDestroyMessage);
	if (!pss->ring) {
		return -1;
	}
	pss->tail = 0;
	return 0;
}

void IZ_WSClientOnClose(struct lws* wsi) {
	IZ_WSClientVHostData* vhd = (IZ_WSClientVHostData*) lws_protocol_vh_priv_get(
		lws_get_vhost(wsi),
		lws_get_protocol(wsi)
	);

	vhd->client_wsi = NULL;
	vhd->established = false;
	lws_sul_schedule(
		vhd->context,
		0,
		&vhd->sul,
		IZ_WSClientAttemptConnect,
		LWS_US_PER_SEC
	);
}

IZ_ProcedureResult IZ_WSClientWritable(struct lws* wsi) {
	IZ_WSClientVHostData* vhd = (IZ_WSClientVHostData*) lws_protocol_vh_priv_get(
		lws_get_vhost(wsi),
		lws_get_protocol(wsi)
	);

	const IZ_WebsocketMessage* pmsg = lws_ring_get_element(vhd->ring, &vhd->tail);
	if (!pmsg) {
		return 0;
	}

	/* notice we allowed for LWS_PRE in the payload already */
	i32 m = lws_write(
		wsi,
		((unsigned char*) pmsg->payload) + LWS_PRE,
		pmsg->len,
		pmsg->binary ? LWS_WRITE_BINARY : LWS_WRITE_TEXT
	);

	if (m < (i32)pmsg->len) {
		lwsl_err("ERROR %d writing to ws socket\n", m);
		return -1;
	}

	lws_ring_consume_single_tail(vhd->ring, &vhd->tail, 1);

	/* more to do for us? */
	if (lws_ring_get_element(vhd->ring, &vhd->tail)) {
		/* come back as soon as we can write more */
		lws_callback_on_writable(wsi);
	}

	return 0;
}

void IZ_WSClientOnReceive(struct lws* wsi, IZ_WSClientSessionData* pss, void* in, size_t len) {
	lwsl_user("LWS_CALLBACK_CLIENT_RECEIVE: %4d (rpp %5d, first %d, last %d, bin %d)\n",
						(int)len, (int)lws_remaining_packet_payload(wsi),
						lws_is_first_fragment(wsi),
						lws_is_final_fragment(wsi),
						lws_frame_is_binary(wsi));

	// lwsl_hexdump_notice(in, len);
	IZ_WebsocketMessage amsg;
	amsg.first = (char)lws_is_first_fragment(wsi);
	amsg.final = (char)lws_is_final_fragment(wsi);
	amsg.binary = (char)lws_frame_is_binary(wsi);
	i32 n = (i32) lws_ring_get_count_free_elements(pss->ring);
	if (!n) {
		lwsl_user("dropping!\n");
		return;
	}

	amsg.len = len;
	/* notice we over-allocate by LWS_PRE */
	amsg.payload = malloc(LWS_PRE + len);
	if (!amsg.payload) {
		lwsl_user("OOM: dropping\n");
		return;
	}

	memcpy((char *)amsg.payload + LWS_PRE, in, len);
	IZ_WSClientVHostData* vhd = (IZ_WSClientVHostData*) lws_protocol_vh_priv_get(
		lws_get_vhost(wsi),
		lws_get_protocol(wsi)
	);

	IZ_App* app = (IZ_App*) vhd->app;
	if (amsg.binary) {
		IZ_AppHandleNetworkingInboundBinaryEvents(app, in, len);
	} else {
		IZ_AppHandleNetworkingInboundTextEvents(app, in, len);
	}

	if (!lws_ring_insert(pss->ring, &amsg, 1)) {
		IZ_WebsocketDestroyMessage(&amsg);
		lwsl_user("dropping!\n");
		return;
	}
	lws_callback_on_writable(wsi);

	if (!pss->flow_controlled && n < 3) {
		pss->flow_controlled = 1;
		lws_rx_flow_control(wsi, 0);
	}
}

IZ_ProcedureResult IZ_WSClientCallback(
	struct lws* wsi,
	enum lws_callback_reasons reason,
	void* user,
	void* in,
	size_t len
) {
	switch (reason) {
		case LWS_CALLBACK_PROTOCOL_INIT:
			return IZ_WSClientProtocolInitialize(wsi, in);
		case LWS_CALLBACK_PROTOCOL_DESTROY:
			IZ_WSClientProtocolTeardown(wsi);
			break;
		case LWS_CALLBACK_CLIENT_CONNECTION_ERROR:
			IZ_WSClientConnectionError(wsi, in);
			break;
		case LWS_CALLBACK_CLIENT_ESTABLISHED:
			return IZ_WSClientOnOpen(wsi, user);
		case LWS_CALLBACK_CLIENT_CLOSED:
			IZ_WSClientOnClose(wsi);
			break;
		case LWS_CALLBACK_CLIENT_RECEIVE:
			IZ_WSClientOnReceive(wsi, user, in, len);
			break;
		case LWS_CALLBACK_CLIENT_WRITEABLE:
			return IZ_WSClientWritable(wsi);
		default:
			break;
	}

	return lws_callback_http_dummy(wsi, reason, user, in, len);
}
